require_relative 'pb/protomodel'
require 'socket'
require 'timeout'
require 'monitor'

class ProtorClient
  include MonitorMixin

  attr_reader :config

  DEFAULT_BUCKET = [0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1, 2.5, 5, 10].freeze

  DEFAULT_CONFIG = {
    host: 'localhost',
    port: 8080,
  }

  def initialize(&block)
    super(&block)

    @families = []
    @config = DEFAULT_CONFIG.dup
    @conn = nil
    yield(config) if block_given?
  end 

  def counter(metric_name, value, labels = {})
    safely do 
      @families << Protomodel::Sample.new(:service => config[:service], :name => metric_name, :kind => "c", :value => value, :label => labels)
    end
    logger.debug("add counter") if logger
  end

  def gauge(metric_name, value, labels = {})
    safely do
      @families << Protomodel::Sample.new(:service => config[:service], :name => metric_name, :kind => "g", :value => value, :label => labels)
    end
    logger.debug("add gauge") if logger
  end

  def histogram(metric_name, value, buckets = DEFAULT_BUCKET, labels = {})
    safely do
      @families << Protomodel::Sample.new(:service => config[:service], :name => metric_name, :kind => "h", :value => value, :label => labels, :histogramDef => buckets)
    end
    logger.debug("add histogram") if logger
  end

  def histogram_linear(metric_name, value, buckets, labels = {})
    safely do
      @families << Protomodel::Sample.new(:service => config[:service], :name => metric_name, :kind => "hl", :value => value, :label => labels, :histogramDef => buckets)
    end
    logger.debug("add histogram linear") if logger
  end

  def publish
    safely do 
      return unless @families.length
      openconn unless @conn
      msg = Protomodel::Samples.new(:array => @families)
      encoded = Protomodel::Samples.encode(msg)
      begin 
        timeout(5) do
          @conn.write(encoded)
          reset if @conn.gets
        end
      rescue Exception => err # second attempt
        openconn
        begin
          timeout(5) do
            @conn.write(encoded)
            reset if @conn.gets
          end
        rescue Exception => err #raise error and put on logger
          logger.error(err) if logger
          raise err unless silent
          return
        end
      end
    end
    logger.debug("publish") if logger
  end

  def closeconn
    @conn.close
    @conn = nil
  end

  private

  def reset
    @families = []
  end

  def safely(&block)
    synchronize(&block)
  rescue StandardError => err
    logger.error(err) if logger
    raise err unless silent
    return
  end

  def openconn
    begin 
      @conn = TCPSocket.open(config[:host], config[:port])
    rescue StandardError => err
      logger.error(err) if logger
      raise err unless silent
      return
    end
  end

  def logger
    config[:logger]
  end

  def silent
    config[:silent]
  end

end
